/*
 * Copyright (C) 2007, Daniel Borges (MsK`) - msk@lywenn.eu.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   3. Neither Daniel Borges nor the names of its contributors may be used
 *	to endorse or promote products derived from this software without
 *	specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _MTPX_H_
#define _MTPX_H_

#include <nds.h>

/*
 * mtpx : Sprite Multiplexer
 *
 * Author  : Daniel Borges (MsK`)
 * Version : 1.0
 *
 *   This code can display hundreds of sprites by updating the OAM while the
 *   screen is rendered. It uses an algorithm that divides the screen in big
 *   rows and count sprites visible in each row, then, at display time,
 *   removes the sprites not visible anymore and add next visible sprites.
 *
 *   - max_spr is the maximum number of sprites displayed.
 *   - hw_spr is the number of sprites actually used in the 2D engine.
 *   - blocksize is the size of each row. (in lines)
 *       half the height of the sprites that will be used with mtpx is
 *       a good value.
 *   - oam is a pointer to where mtpx will place visible sprites of
 *     main screen.
 *       if oam is null, mtpx will not multiplex on this screen.
 *   - oam_sub is a pointer to where mtpx will place visible sprites of
 *     sub screen.
 *       if oam_sub is null, mtpx will not multiplex on this screen.
 *
 *  !! Attention !!
 *  When too many sprites are displayed on a row, some sprites will not be
 *  displayed fully, as it will not have enough "hardware" sprites to render
 *  all of them.
 *
 *  Nintendo DS DMA Limits :
 *  The Horizontal Blank interrupt is a short period to update both OAMs.
 *  Don't try to multiplex on both screens with too many hardware sprites,
 *  if you do, first line of each sprite-block may not be rendered correctly.
 *
 *  This code was *highly* inspired by Takayama Fumihiko's job on BulletGBA.
 *  ( http://gba.pqrs.org/~tekezo/gba/doc/sprite-multiplexer/index.html )
 *
 */

template <u16 max_spr, u8 hw_spr, u8 blocksize, u32 oam, u32 oam_sub>
class mtpx {

    static const int NB_BLOCKS = SCREEN_HEIGHT / blocksize + 1;
    static const int DOUBLEBUFFER = 2;
    static const int BOTHSCREENS = 2;
    static const int HWSPRCOPYSIZE = hw_spr * sizeof(SpriteEntry);

    // Number of sprites registered (per screen)
    static u16 registered[BOTHSCREENS];
    // Number of sprites registered (per row, per screen)
    static u16 regist[BOTHSCREENS][NB_BLOCKS];
    // Sprites registered (per screen)
    static SpriteEntry * added[BOTHSCREENS][max_spr];

    // Sprites semi-ordered (per buffer, per screen)
    static SpriteEntry sprites[DOUBLEBUFFER][BOTHSCREENS][max_spr];
    // Pointers to sprites that point to the first sprite visible
    // of each row (per buffer, per screen, per row)
    static SpriteEntry * start[DOUBLEBUFFER][BOTHSCREENS][NB_BLOCKS];
    // Number of the buffer currently displayed
    static u8 displayed;

public:

    typedef enum { MAIN = 1, SUB = 0 } screen;

    /* 
     * Initialize multiplexer.
     *   Puts a zero sprite frame in and sets IRQ handlers, if requested.
     *   IRQ Handlers :
     *	   frame_update as a HBL interrupt
     */
    static void setup(bool set_hbl);

    /*
     * Prepare next frame.
     *   Clears data.
     */
    static void init_frame(void);
    /*
     * Add a sprite.
     *   Returns 0 if the sprite is refused, 1 otherwise.
     */
    static int add(screen scr, const SpriteEntry * spr);
    /*
     * Add lots of sprites.
     *   Way much faster than adding sprite by sprite...
     */
    static void add(screen scr, const SpriteEntry * sprs, u16 count);
    /*
     * Compile sprites
     *   Make them ready for display phase.
     */
    static void compile(void);
    /*
     * Swap buffers
     */
    inline static void swap(void) { displayed = ! displayed; }

    /*
     * Frame update
     *   Call it in your Horizontal Blank Interrupt.
     */
    static void frame_update(void);

    /*
     * Get the number of sprites registered
     */
    static u16 get_registered(screen scr) { return registered[scr]; };

};

typedef class mtpx<1024, 64, 4, 0x07000000, 0x07000400> sprmtpx;

#endif // _MTPX_H_
